"""
A module for instantiating sparse tables with log probabilities.
"""

# System imports
import copy
import operator
import warnings

# Third-party imports
import numpy as np
from numpy import unravel_index
from scipy import special
from itertools import product

# Local imports
from veroku.factors._factor import Factor
from veroku.factors._factor_template import FactorTemplate

# special operator rules
LOG_SUBTRACT_CANCEL_RULES = {(-np.inf, operator.sub, -np.inf): -np.inf}
LOG_SUBTRACT_KL_RULES = {(-np.inf, operator.sub, -np.inf): 0.0}


# TODO: consider removing some unused functions

# TODO: Currently variable values should be the same as numpy indices and therefore start at 0 with increments of 1.
#       We probably want to relax this requirement.


class Categorical(Factor):
    """
    A class for instantiating sparse tables with log probabilities.
    """

    def __init__(self, var_names, cardinalities=None, probs_table=None, log_probs_tensor=None):
        """
        Construct a SparseLogTable. Either log_probs_table or probs_table should be supplied.

        :param var_names: The variable names.
        :type var_names: str list
        :param log_probs_table: A dictionary with assignments (tuples) as keys and probabilities as values.
            Missing assignments are assumed to have zero probability.
        :type
        :param log_probs_table: A dictionary with assignments (tuples) as keys and log probabilities as values.
            Missing assignments are assumed to have -inf log-probability (zero probability).
        Example:
            >>> var_names = ['rain','slip']
            >>> probs_table = {(0,0):0.8,
            >>>                (0,1):0.2,
            >>>                (1,0):0.4,
            >>>                (1,1):0.6}
            >>>var_cardinalities = [2,2]
            >>>table = Categorical(log_probs_table=log_probs_table,
            >>>                    var_names=var_names,
            >>>                    cardinalities=var_cardinalities)
        """
        # TODO: add check that assignment lengths are consistent with var_names
        # TODO: add check that cardinalities are consistent with assignments
        super().__init__(var_names=var_names)

        if cardinalities is None:
            if log_probs_tensor is None:
                raise ValueError('numpy array type log_probs_tensor is expected cardinalities are not supplied. Alternatively, provide cardinalities with dict type probs_table.')
            cardinalities = log_probs_tensor.shape
        elif len(cardinalities) != len(var_names):
            raise ValueError('The cardinalities and var_names lists should be the same length.')
        if (log_probs_tensor is None) and (probs_table is None):
            raise ValueError('Either log_probs_tensor or probs_table must be specified')
        if log_probs_tensor is None:
            probs_tensor = np.zeros(shape=cardinalities)
            for assignment, prob in probs_table.items():
                try:
                    probs_tensor[assignment] = prob
                except IndexError:
                    error_message = f'assignment {assignment} is not consistent with the cardinalities ({cardinalities}) provided'
                    raise IndexError(error_message)
            with warnings.catch_warnings():
                # prevents 'divide by zero encountered in log' from displaying
                warnings.simplefilter("ignore")
                self.log_probs_tensor = np.log(probs_tensor)
        else:
            if len(log_probs_tensor.shape) != len(var_names):
                raise ValueError('log_probs_tensor should have dimension equal to the number of variables.')
            self.log_probs_tensor = log_probs_tensor.copy()
        # TODO: remove this
        self.var_cards = dict(zip(var_names, cardinalities))
        self.cardinalities = cardinalities

    def reorder(self, new_var_names_order):
        """
        Reorder categorical table variables to a new order and reorder the associated probabilities
        accordingly.
        :param new_var_names_order: The new variable order.
        :type new_var_names_order: str list
        :return: The factor with new order.

        Example:
            old_variable_order = [a, b]
            new_variable_order = [b, a]

            a b P(a,b)  return    b a P(a,b)
            0 0  pa0b0            0 0  pa0b0
            0 1  pa0b1            0 1  pa1b0
            1 0  pa1b0            1 0  pa0b1
            1 1  pa1b1            1 1  pa1b1
        """
        if new_var_names_order == self.var_names:
            return self.copy()
        if set(new_var_names_order) != set(self.var_names):
            raise ValueError('The new_var_names_order set must be the same as the factor variables.')
        vars_new_order_indices = [self.var_names.index(v) for v in new_var_names_order]
        log_probs_tensor = self.log_probs_tensor.transpose(vars_new_order_indices)
        return Categorical(var_names=new_var_names_order, log_probs_tensor=log_probs_tensor)

    def equals(self, factor):
        """
        Check if this factor is the same as another factor.

        :param factor: The other factor to compare to.
        :type factor: Categorical
        :return: The result of the comparison.
        :rtype: bool
        """
        if not isinstance(factor, Categorical):
            raise ValueError(f'factor must be of Categorical type but has type {type(factor)}')
        if set(self.var_names) != set(factor.var_names):
            return False
        factor_copy = factor.reorder(self.var_names)
        if not np.allclose(self.log_probs_tensor, factor_copy.log_probs_tensor):
            return False
        return True

    def copy(self):
        """
        Make a copy of this factor.
        :return: The copy of this factor.
        :rtype: Categorical
        """
        return Categorical(var_names=self.var_names.copy(),
                           log_probs_tensor=self.log_probs_tensor.copy())

    @staticmethod
    def _get_shared_order_smaller_vars(smaller_vars, larger_vars):
        """
        larger_vars = ['a', 'c', 'd', 'b']
        smaller_vars = ['c', 'e', 'b']
        return ['c', 'b']
        """
        shared_vars = [v for v in smaller_vars if v in larger_vars]
        remaining_smaller_vars = list(set(larger_vars) - set(shared_vars))
        smaller_vars_new_order = shared_vars + remaining_smaller_vars
        return smaller_vars_new_order

    @staticmethod
    def _intersection_has_same_order(larger_vars, smaller_vars):
        """
        Check if the intersection of two lists has the same order in both lists.
        Will return true if either list is empty? SHOULD THIS BE THE CASE?
        """
        indices_of_smaller_in_larger = [larger_vars.index(v) for v in smaller_vars if v in larger_vars]
        if sorted(indices_of_smaller_in_larger) == indices_of_smaller_in_larger:
            return True
        return False

    @staticmethod
    def tensor_operation(tensor_a, tensor_b, tensor_a_var_names, tensor_b_var_names, func):
        """
        Apply a element wise function to two tensors with named indices.
        :param numpy.ndarray tensor_a: The first tensor.
        :param numpy.ndarray tensor_b: The second tensor.
        :param tensor_a_var_names: The first tensor's variable names corresponding to the indices.
        :type tensor_a_var_names: string list
        :param tensor_b_var_names: The second tensor's variable names corresponding to the indices.
        :type tensor_b_var_names: string list
        :param func: The function to apply elementwise to the two tensors (func(a,b))
        :return: The resulting tensor and corresponding variable names
        :rtype: string list
        """
        if tensor_a_var_names == tensor_b_var_names:
            result_tensor = func(tensor_a, tensor_b)
            return result_tensor, tensor_a_var_names
        common_vars = set(tensor_a_var_names).intersection(tensor_b_var_names)

        remaining_f1_vars = [v for v in tensor_a_var_names if v not in common_vars]
        remaining_f2_vars = [v for v in tensor_b_var_names if v not in common_vars]
        # sort according to f1_vars so that only f2_tensor can be transposed if sets are the same
        common_vars = [v for v in tensor_a_var_names if v in common_vars]
        f1_vars_new_order = [tensor_a_var_names.index(v) for v in remaining_f1_vars]
        f1_vars_new_order += [tensor_a_var_names.index(v) for v in common_vars]

        f1_tensor_new_shape = tensor_a.transpose(f1_vars_new_order)
        f1_dim_expansion_axis = list(range(len(remaining_f2_vars)))
        f1_tensor_std_shape = np.expand_dims(f1_tensor_new_shape, axis=f1_dim_expansion_axis)
        #  now f1 has var order / dim order: [[1]*num_f2_vars, rest_f1_vars_dims, common_vars]

        f2_vars_new_order = [tensor_b_var_names.index(v) for v in remaining_f2_vars]
        f2_vars_new_order += [tensor_b_var_names.index(v) for v in common_vars]
        f2_tensor_new_shape = tensor_b.transpose(f2_vars_new_order)
        # now f2 has var order / dim order:  [common_vars_dim, [1]*num_f2_vars]

        f2_dim_expansion_axis = list(range(len(remaining_f2_vars), len(remaining_f2_vars) + len(remaining_f1_vars)))
        f2_tensor_std_shape = np.expand_dims(f2_tensor_new_shape, axis=f2_dim_expansion_axis)
        # now f1 has var order / dim order: [f1_vars_dims, [1]*num_f1_vars] =
        # [common_vars, rest_f1_vars_dims, [1]*num_f2_vars]
        result_vars = remaining_f2_vars + remaining_f1_vars + common_vars
        result_tensor = func(f1_tensor_std_shape, f2_tensor_std_shape)
        return result_tensor, result_vars

    # TODO: change back to log form
    def marginalize(self, vrs, keep=False):
        """
        Sum out variables from this factor.
        :param vrs: (list) a subset of variables in the factor's scope
        :param keep: Whether to keep or sum out vrs
        :return: The resulting factor.
        :rtype: Categorical
        """

        vars_to_keep = super().get_marginal_vars(vrs, keep)
        vars_to_sum_out = [v for v in self.var_names if v not in vars_to_keep]
        indices_to_sum_out = [self.var_names.index(v) for v in vars_to_sum_out]
        result_tensor = np.apply_over_axes(special.logsumexp, self.log_probs_tensor, axes=indices_to_sum_out)
        return Categorical(var_names=vars_to_keep, log_probs_tensor=np.squeeze(result_tensor))

    def reduce(self, vrs, values):
        """
        Observe variables to have certain values and return reduced table.
        :param vrs: (list) The variables.
        :param values: (tuple or list) Their values
        :return: The resulting factor.
        :rtype: Categorical
        """

        vars_unobserved = [v for v in self.var_names if v not in vrs]
        obs_dict = dict(zip(vrs, values))

        reduced_tensor_indexing = tuple([obs_dict[v] if v in obs_dict else slice(None) for v in self.var_names])
        result_table = self.log_probs_tensor[reduced_tensor_indexing]
        return Categorical(var_names=vars_unobserved, log_probs_tensor=result_table)

    def _assert_consistent_cardinalities(self, factor):
        """
        Assert that the variable cardinalities are consistent between two factors.

        :param factor:
        """
        for var in self.var_names:
            if var in factor.var_cards:
                error_msg = f'Error: inconsistent variable cardinalities: {factor.var_cards}, {self.var_cards}'
                assert self.var_cards[var] == factor.var_cards[var], error_msg

    def multiply(self, factor):
        """
        Multiply this factor with another factor and return the result.

        :param factor: The factor to multiply with.
        :type factor: Categorical
        :return: The factor product.
        :rtype: Categorical
        """
        if not isinstance(factor, Categorical):
            raise ValueError(f'factor must be of SparseLogTable type but has type {type(factor)}')
        self._assert_consistent_cardinalities(factor)
        result_tensor, result_vars = self.tensor_operation(self.log_probs_tensor, factor.log_probs_tensor,
                                                           self.var_names, factor.var_names, operator.add)
        return Categorical(var_names=result_vars, log_probs_tensor=result_tensor)

    def cancel(self, factor):
        """
        Almost like divide, but with a special rule that ensures that division of zeros by zeros results in zeros. When
        categorical message factors with zero probabilities are used in Belief Update algorithms, multiplication by the
        zero probabilities cause zeros in the cluster potentials, when these messages need to be divided out again, this
        results in 0/0 operations. Since, in such cases, we know (from the information in the message) that the
        probability value should be zero, it makes sense to set the result of 0/0 operations to 0 in these cases.
        :param factor: The factor to divide by.
        :type factor: Categorical
        :return: The factor quotient.
        :rtype: Categorical
        """
        augmented_factor_tensor = factor.log_probs_tensor.copy()
        # when cancelling out a factor
        special_case_indices = np.where((augmented_factor_tensor == -np.inf) & (self.log_probs_tensor == -np.inf))

        indices_are_empty = [a.size == 0 for a in special_case_indices]
        if not any(indices_are_empty):
            augmented_factor_tensor[special_case_indices] = np.float(0.0)
        elif not all(indices_are_empty):
            print('augmented_factor_tensor: ')
            print(augmented_factor_tensor)
            print('self.log_probs_tensor: ')
            print(self.log_probs_tensor)
            raise ValueError('something strange happened.')
        result_tensor, result_vars = self.tensor_operation(self.log_probs_tensor, augmented_factor_tensor,
                                                           self.var_names, factor.var_names, operator.sub)
        return Categorical(var_names=result_vars, log_probs_tensor=result_tensor)

    def divide(self, factor, special_rules=None):
        """
        Divide this factor by another factor and return the result.
        :param factor: The factor to divide by.
        :type factor: Categorical
        :param special_rules: Any special rules to apply for specific values of the left and right variables.
            For example: {(left_var_val, right_var_val): result, ...}
        :type special_rules: dict
        :return: The factor quotient.
        :rtype: Categorical
        """
        if not isinstance(factor, Categorical):
            raise ValueError(f'factor must be of SparseLogTable type but has type {type(factor)}')
        self._assert_consistent_cardinalities(factor)
        result_tensor, result_vars = self.tensor_operation(self.log_probs_tensor, factor.log_probs_tensor,
                                                           self.var_names, factor.var_names, operator.sub)
        return Categorical(var_names=result_vars, log_probs_tensor=result_tensor)

    def argmax(self):
        """
        Get the Categorical assignment (vector value) that maximises the factor potential.

        :return: The argmax assignment.
        :rtype: int list
        """

        argmax_index = unravel_index(self.log_probs_tensor.argmax(), self.log_probs_tensor.shape)
        return argmax_index

    def normalize(self):
        """
        Return a normalized copy of the factor.
        :return: The normalized factor.
        """

        factor_copy = self.copy()
        logz = special.logsumexp(self.log_probs_tensor)
        factor_copy.log_probs_tensor -= logz
        return factor_copy

    @property
    def is_vacuous(self):
        """
        Check if this factor is vacuous (i.e uniform).
        :return: Whether the factor is vacuous or not.
        :rtype: bool
        """
        if self.distance_from_vacuous() < 1e-10:
            return True
        return False

    def kl_divergence(self, factor, normalize_factor=True):
        """
        Get the KL-divergence D_KL(P||Q) = D_KL(self||factor) between a normalized version of this factor and another factor.
        Reference https://infoscience.epfl.ch/record/174055/files/durrieuThiranKelly_kldiv_icassp2012_R1.pdf, page 1.

        :param factor: The other factor
        :type factor: Gaussian
        :param normalize_factor: Whether or not to normalize the other factor before computing the KL-divergence.
        :type normalize_factor: bool
        :return: The Kullback-Leibler divergence
        :rtype: float
        """
        normalized_self = self.normalize()
        log_p = normalized_self.log_probs_tensor
        factor_ = factor
        if normalize_factor:
            factor_ = factor.normalize()
        log_q = factor_.log_probs_tensor
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            kl_array = np.where(log_p != -np.inf, np.exp(log_p) * (log_p - log_q), 0.0)
        kld = np.sum(kl_array)
        if kld < 0.0:
            if np.isclose(kld, 0.0, atol=1e-5):
                #  this is fine (numerical error)
                return 0.0
            print('\nnormalize_factor = ', normalize_factor)
            print('self = ')
            self.show()
            print('normalized_self = ')
            normalized_self.show()
            print('\nfactor = ')
            factor_.show()
            raise ValueError(f'Negative KLD: {kld}')
        return kld

    def distance_from_vacuous(self):
        """
        Get the Kullback-Leibler (KL) divergence between this factor and a uniform copy of it.

        :return: The KL divergence.
        :rtype: float
        """
        # make uniform copy
        uniform_log_prob = -np.log(np.product(self.log_probs_tensor.shape))
        uniform_log_tensor = np.ones(self.log_probs_tensor.shape) * uniform_log_prob
        uniform_factor = Categorical(var_names=self.var_names, log_probs_tensor=uniform_log_tensor)
        kl = self.kl_divergence(uniform_factor, normalize_factor=False)
        if kl < 0.0:
            raise ValueError(f"kl ({kl}) < 0.0")
            self.show()
        return kl

    def potential(self, vrs, assignment):
        """
        Get the value of the factor for a specific assignment.

        :param assignment: The assignment
        :return: The value
        """
        assert set(vrs) == set(self.var_names), 'variables (vrs) do not match factor variables.'
        obs_dict = dict(zip(vrs, assignment))
        obs_tensor_indexing = tuple([obs_dict[v] if v in obs_dict else slice(None) for v in self.var_names])
        return self.log_probs_tensor[obs_tensor_indexing]

    def show(self, exp_log_probs=True):
        """
        Print the factor.

        :param exp_log_probs: Whether or no to exponentiate the log probabilities (to display probabilities instead of
        log-probabilities)
        :type exp_log_probs: bool
        """
        prob_string = 'log(prob)'
        if exp_log_probs:
            prob_string = 'prob'
        print(self.var_names, ' ', prob_string)
        for assignment in np.ndindex(self.log_probs_tensor.shape):
            prob = self.log_probs_tensor[assignment]
            if exp_log_probs:
                prob = np.exp(prob)
            print(assignment, ' ', prob)


class CategoricalTemplate(FactorTemplate):

    def __init__(self, log_probs, var_templates):
        """
        Create a Categorical factor template.

        :param probs_table: The log_probs_table that specifies the assignments and values for the template.
        :type probs_table: tuple:float dict
        :param var_templates: A list of formattable strings.
        :type var_templates: str list

        log_probs_table example:
        {(0, 0): 0.1,
         (0, 1): 0.3,
         (1, 0): 0.1,
         (1, 1): 0.5}
        """
        # TODO: Complete and improve docstring.
        super().__init__(var_templates=var_templates)
        self.log_probs_table = copy.deepcopy(log_probs)

    def make_factor(self, format_dict=None, var_names=None):
        """
        Make a factor with var_templates formatted by format_dict to create specific var names.

        :param format_dict: The dictionary to be used to format the var_templates strings.
        :type format_dict: str dict
        :return: The instantiated factor.
        :rtype: Categorical
        """
        if format_dict is not None:
            assert var_names is None
            var_names = [vt.format(**format_dict) for vt in self._var_templates]
        return Categorical(probs_table=copy.deepcopy(self.log_probs),
                           var_names=var_names, cardinalities=self.var_cards.values())



from IPython.core.display import Image, display
from graphviz import Graph
import networkx as nx
import numpy as np
from tqdm.auto import tqdm
import time
import pandas as pd
from graphviz import Source

from veroku._cg_helpers._cluster import Cluster
import veroku._cg_helpers._animation as cg_animation
from veroku.factors._factor_utils import get_subset_evidence
import matplotlib.pyplot as plt
from veroku.factors.gaussian import Gaussian


# TODO: optimise _pass_message
# TODO: improve sepsets selection for less loopiness
# TODO: Optimisation: messages from clusters that did not receive any new messages in the previous round, do not need
#  new messages calculated.


def _evidence_reduce_factors(factors, evidence):
    """
    Observe relevant evidence for each factor.
    :param factors:
    :param evidence:
    :return:
    """
    reduced_factors = []
    for i, factor in enumerate(factors):
        if evidence is not None:
            vrs, values = get_subset_evidence(all_evidence_dict=evidence,
                                              subset_vars=factor.var_names)
            if len(vrs) > 0:
                factor = factor.reduce(vrs, values)
        reduced_factors.append(factor.copy())
    return reduced_factors


def make_factor_name(factor):
    return str(factor.var_names).replace("'", '')


def _absorb_subset_factors(factors):
    """
    Absorb any factors that has a scope that is a subset of another factor into such a factor.
    :param factors:
    :return:
    """
    factors_absorbtion_dict = {i: [] for i in range(len(factors))}
    final_graph_cluster_factors = []
    # factors: possibly smaller list of factors after factors which have a scope
    # that is a subset of another factor have been absorbed by the larger one.
    # print(f"Info: Absorbing subset factors. (initial number of factors: {len(factors)})")
    factor_processed_mask = [0] * len(factors)
    for i, factor_i in enumerate(factors):
        if not factor_processed_mask[i]:
            factor_product = factor_i.copy()
            for j, factor_j in enumerate(factors):
                if i != j:
                    if not factor_processed_mask[j]:
                        if set(factor_j.var_names) < set(factor_product.var_names):
                            # print(f'Absorbing {factor_j.var_names} into{factor_product.var_names}')
                            # factors_absorbtion_dict[make_factor_name(factors[i])].append(make_factor_name(factors[j]))
                            try:
                                factor_product = factor_product.multiply(factor_j)
                                factors_absorbtion_dict[i].append(j)
                                factor_processed_mask[j] = 1
                                factor_processed_mask[i] = 1
                            except NotImplementedError:
                                print(f'Warning: could not multiply {type(factor_product)} with {type(factor_j)} (Not Implemented)')
            if factor_processed_mask[i]:
                final_graph_cluster_factors.append(factor_product)
    for i, factor_i in enumerate(factors):  # add remaining factors
        if not factor_processed_mask[i]:
            factor_processed_mask[i] = 1
            final_graph_cluster_factors.append(factor_i)
    assert all(factor_processed_mask), 'Error: Some factors where not included during variable subset processing.'
    return final_graph_cluster_factors, _make_subset_factor_df(factors_absorbtion_dict)


def _make_subset_factor_df(subset_dict):
    """
    Make a ...
    :param subset_dict: (dict) A dictionary mapping factors to factors that have subset scopes
    :return:
    """
    keys = list(subset_dict.keys())
    values = list(subset_dict.values())
    assert(len(keys) == len(values))
    # TODO: This raises different list lengths warning (see below). Investigate this.
    #   VisibleDeprecationWarning: Creating an ndarray from ragged nested sequences...
    data = np.array([keys, values]).T
    df = pd.DataFrame(columns=['factor_index', 'subfactor_indices'],
                      data=data)
    df['num_subfactors'] = df['subfactor_indices'].apply(lambda x: len(x))
    df.sort_values(by='num_subfactors', inplace=True, ascending=False)
    return df


class ClusterGraph(object):
    
    def __init__(self, factors, evidence=None, special_evidence=dict(), make_animation_gif=False, debug=False,
                 disable_tqdm=False, verbose=False):
        """
        Construct a Cluster graph from a list of factors
        :param factors: (list of factors) The factors to construct the graph from
        :param evidence: (dict) evidence dictionary (mapping variable names to values) that should be used to reduce
            factors before building the cluster graph.
        :param special_evidence: (dict) evidence dictionary (mapping variable names to values) that should be used in
            the calculation of messages, and not to reduce factors. This allows factor approximations - such as the
            non-linear Gaussian to be iteratively refined.
        """

        self.full_mp_iters = 0
        self.debug = debug
        self.debug_passed_message_factors = []
        self.num_messages_passed = 0
        self.make_animation_gif = make_animation_gif
        self.special_evidence = special_evidence
        self.disable_tqdm = disable_tqdm
        self.message_passing_index_strings = []
        self.last_passed_message_factors_dict = dict()
        self.verbose = verbose
        # new
        self.last_sent_message_dict = {}  # {(rec_cluster_id1, rec_cluster_id1): msg1, ...}

        self.sync_message_passing_max_distances = None
        self.sync_message_passing_max_distances = None

        all_evidence_vars = set(self.special_evidence.keys())
        if evidence is not None:
            evidence_vars = set(evidence.keys())
            all_evidence_vars = all_evidence_vars.union(evidence_vars)

        prev_time = time.time()
        all_factors_copy = _evidence_reduce_factors(factors, evidence)
        self._conditional_print(f'Debug: Copy factors and reduce evidence time duration: {time.time() - prev_time}')

        prev_time = time.time()
        final_graph_cluster_factors, absorbtion_df = _absorb_subset_factors(all_factors_copy)

        self._conditional_print(f'Debug: Absorbing subset factors time duration: {time.time() - prev_time}')
        self._conditional_print(f"Info: Factor subset factors. (reduced number of factors: {len(final_graph_cluster_factors)}")

        prev_time = time.time()
        clusters = [Cluster(factor, cluster_name_prefix=f'c{i}#') for i, factor in
                    enumerate(final_graph_cluster_factors)]
        self._non_rip_sepsets = {}

        self._conditional_print('Info: Calculating sepsets')

        for i in tqdm(range(len(clusters)), disable=self.disable_tqdm):
            vars_i = clusters[i].var_names
            for j in range(i + 1, len(clusters)):
                vars_j = clusters[j].var_names
                sepset = set(vars_j).intersection(set(vars_i)) - all_evidence_vars
                self._non_rip_sepsets[(i, j)] = sepset
                self._non_rip_sepsets[(j, i)] = sepset
                if len(sepset) > 0:
                    clusters[i].add_neighbour(clusters[j])
        self._clusters = clusters

        #TODO: copy here?
        self._cluster_id_cluster_dict = {c._cluster_id: c for c in self._clusters}

        self._conditional_print(f'Debug: Calculating sepsets time duration: {time.time() - prev_time}')
        self._conditional_print(f'Debug: number of clusters: {len(self._clusters)} (should be {len(final_graph_cluster_factors)})')

        prev_time = time.time()

        self._conditional_print(f'Debug: building graph')
        self._build_graph()
        self._conditional_print(f'Debug: build graph time duration: {time.time() - prev_time}')

        # TODO: consolodate these two, if possible
        self.message_passing_log_df = None
        self.message_passing_animation_frames = []

    def _build_graph(self, rip_sepsets_dict=None):
        """
        Add the cluster sepsets, graphviz graph and animation graph (for message_passing visualisation).
        """
        # Check for non-unique cluster_ids (This should never be the case)
        cluster_ids = self._get_cluster_ids()
        if len(set(cluster_ids)) != len(cluster_ids):
            raise ValueError(f'Non-unique cluster ids: {cluster_ids}')
        # self.graph_animation = Animation()
        self._conditional_print('Info: Building graph.')
        self._graph = Graph(format='png')
        if rip_sepsets_dict is None:
            rip_sepsets_dict = self._get_running_intersection_sepsets()

        self._conditional_print(f'Debug: number of clusters: {len(self._clusters)}')
        for i in tqdm(range(len(self._clusters)), disable=self.disable_tqdm):
            # TODO: add cleaner solution?
            self._clusters[i].remove_all_neighbours()

        for i in tqdm(range(len(self._clusters)), disable=self.disable_tqdm):

            node_i_name = self._clusters[i]._cluster_id
            self._graph.node(name=node_i_name, label=node_i_name, style='filled', fillcolor='white', color='black')
            for j in range(i + 1, len(self._clusters)):

                if (i, j) in rip_sepsets_dict:
                    sepset = rip_sepsets_dict[(i, j)]
                    assert len(sepset) > 0, 'Error: empty sepset'

                    self._clusters[i].add_neighbour(self._clusters[j], sepset=sepset)
                    self._clusters[j].add_neighbour(self._clusters[i], sepset=sepset)
                    node_j_name = self._clusters[j]._cluster_id
                    sepset_node_label = ','.join(sepset)
                    sepset_node_name = cg_animation.make_sepset_node_name(node_i_name, node_j_name)
                    self._graph.node(name=sepset_node_name, label=sepset_node_label, shape='rectangle')
                    self._graph.edge(node_i_name, sepset_node_name, color='black', penwidth='2.0')
                    self._graph.edge(sepset_node_name, node_j_name, color='black', penwidth='2.0')

    def _conditional_print(self, message):
        if self.verbose:
            print(message)

    def _get_cluster_ids(self):
        """
        Get all the ids for all the clusters.
        :return:
        """
        return [cluster.cluster_id for cluster in self._clusters]

    def plot_message_convergence(self, old=False):
        """
        Plot the the KL-divergence between the messages and their previous instances to indicate the message passing
        convergence.
        :param old: Whether or not to use the older function for synchronous message passing convergence.
        :type old: Bool
        """
        if self.sync_message_passing_max_distances:
            assert self.message_passing_log_df is None, 'Error: it seems bot sync and async message passing was run.'
            plt.plot(self.sync_message_passing_max_distances)
        else:
            if old:
                self._plot_message_convergence_old()
            else:
                self._plot_message_convergence_new()

    def _plot_message_convergence_old(self):
        """
        Plot the message passing convergence over the consecutive iterations.
        """
        # TODO: combine with one below - this one gets messed up by inf values
        message_kld_df = self.message_passing_log_df
        columns_mask = message_kld_df.columns.to_series().str.contains('distance_from_previous')
        message_kld_distances_only_df = message_kld_df[message_kld_df.columns[columns_mask]]
        message_kld_distances_only_df.columns = range(message_kld_distances_only_df.shape[1])

        title = 'KL Divegences between Messages and Previous Iteration Messages'
        ax = message_kld_distances_only_df.plot.box(figsize=[20, 10], title=title)
        ax.title.set_size(20)
        ax.set_ylabel('KLD', fontsize=15)
        ax.set_xlabel('message passing iteration', fontsize=15)

    def _plot_message_convergence_new(self, figsize=[10, 5]):
        # TODO: improve this (inf value workaround is a bit hacky)
        df = self.message_passing_log_df
        kl_cols = [col for col in df.columns if 'distance' in col]
        kl_df = df[kl_cols]

        no_inf_df = kl_df.replace([np.inf], 0) * 2
        max_no_inf = np.max(no_inf_df.values)

        no_inf_df = kl_df.replace([-np.inf], 0) * 2
        min_no_inf = np.min(no_inf_df.values)

        inf_to_max_df = kl_df.replace([np.inf], max_no_inf * 2)
        inf_to_max_min_df = inf_to_max_df.replace([-np.inf], min_no_inf * 2)
        data = inf_to_max_min_df.values
        max_kl_div_per_iteration = np.max(data, axis=0)
        plt.figure(figsize=figsize)
        plt.plot(np.log(max_kl_div_per_iteration))
        plt.title('Message Passing Convergence', fontsize=15)
        plt.ylabel('log max message kld')
        plt.xlabel('message passing iteration')

    def _get_unique_vars(self):
        all_vars = []
        for cluster in self._clusters:
            all_vars += (cluster.var_names)
        unique_vars = list(set(all_vars))
        return unique_vars

    def _get_vars_min_spanning_trees(self):
        all_vars = self._get_unique_vars()
        var_graphs = {var: nx.Graph() for var in all_vars}
        num_clusters = len(self._clusters)
        for i in range(num_clusters):
            for j in range(i + 1, num_clusters):
                sepset = self._non_rip_sepsets[(i, j)]
                for var in sepset:
                    var_graphs[var].add_edge(i, j, weight=1)
        var_spanning_trees = dict()
        for var in all_vars:
            var_spanning_trees[var] = nx.minimum_spanning_tree(var_graphs[var])
        return var_spanning_trees

    def _get_running_intersection_sepsets(self):
        edge_sepset_dict = {}
        unique_vars = self._get_unique_vars()
        min_span_trees = self._get_vars_min_spanning_trees()
        self._conditional_print("Info: Getting unique variable spanning trees.")
        for i in tqdm(range(len(unique_vars)), disable=self.disable_tqdm):
            var = unique_vars[i]
            min_span_tree = min_span_trees[var]
            for edge in min_span_tree.edges():
                if edge in edge_sepset_dict:
                    edge_sepset_dict[edge].append(var)
                else:
                    edge_sepset_dict[edge] = [var]
        return edge_sepset_dict

    def show(self):
        self._graph.render('/tmp/test.gv', view=False)
        image = Image('/tmp/test.gv.png')
        display(image)

    def save_graph_image(self, filename):
        """
        Save image of the graph.
        :param filename: The filename of the file.
        """
        # Source(self._graph, filename="/tmp/test.gv", format="png")
        Source(self._graph, filename=filename, format="png")

    def get_factors(self):
        """
        Get the cluster factor for each cluster in the graph
        :return: A list of cluster factors.
        """
        return [cluster._factor.copy() for cluster in self._clusters]

    def get_marginal(self, vrs):
        """
        Search the graph for a specific variable and get that variables marginal (posterior marginal if process_graph
        has been run previously).
        :return: The marginal
        """
        for cluster in self._clusters:
            if set(vrs) <= set(cluster.var_names):
                factor = cluster._factor.copy()
                evidence_vrs, evidence_values = get_subset_evidence(self.special_evidence, factor.var_names)
                if len(evidence_vrs) > 0:
                    factor = factor.reduce(evidence_vrs, evidence_values)
                marginal = factor.marginalize(vrs, keep=True)
                return marginal
        raise ValueError(f'No cluster with variables containing {vrs}')

    def get_posterior_joint(self):
        """
        Get the posterior joint distribution.
        """
        # TODO: add functionality for efficiently getting a posterior marginal over any subset of variables and replace
        #  the get_marginal function above.
        cluster_product = self._clusters[0]._factor.joint_distribution
        for cluster in self._clusters[1:]:
            cluster_product = cluster_product.multiply(cluster._factor.joint_distribution)
        last_passed_message_factors = list(self.last_passed_message_factors_dict.values())
        if len(last_passed_message_factors) == 0:
            assert self.num_messages_passed == 0
            return cluster_product
        message_product = last_passed_message_factors[0]
        for message_factor in last_passed_message_factors[1:]:
            message_product = message_product.multiply(message_factor)
        joint = cluster_product.cancel(message_product)
        return joint

    def _get_factor(self, cluster_id):
        """
        Get the factor associated with a specific cluster.
        :param cluster_id: The id of the cluster of which the factor will be returned.
        :return: The factor.
        """
        print('number of clusters = ', len(self._clusters))
        for cluster in self._clusters:
            print(f'cluster.cluster_id  {cluster.cluster_id} has type {type(cluster.cluster_id)}')
            if cluster.cluster_id == cluster_id:
                return cluster._factor.copy()
        raise ValueError(f'Could not find cluster with id {cluster_id}')

    def _make_cluster_messages(self, cluster):
        messages = []
        for neighbour_id in cluster.neighbour_ids:
            message = cluster.make_message(neighbour_id, evidence=self.special_evidence)
            messages.append(message)
        return messages

    def _make_all_messages(self):
        """
        Iterate through all clusters and make messages to all the neighbours of a cluster for each of
        the clusters.
        """
        self.full_mp_iters += 1
        all_messages = []

        for cluster in self._clusters:
            cluster_messages = self._make_cluster_messages(cluster)
            all_messages += cluster_messages
        return all_messages

    def _make_clusters_messages_dict(self):
        """
        Iterate through all clusters and make messages to all the neighbours of a cluster for each of
        the clusters.
        :return: A dictionary of cluster_id's mapped to the messages from the respective clusters.
        :rtype: dict
        """
        self.full_mp_iters += 1
        clusters_messages_dict = dict()
        for cluster in self._clusters:
            cluster_messages = self._make_cluster_messages(cluster)
            clusters_messages_dict[cluster.cluster_id] = cluster_messages
        return clusters_messages_dict

    def _get_ranked_message_df(self, previous_message_df=None):
        """
        Get a dataframe containing message objects, their sender and receiver cluster ids, and the distance from their
        previous iterations.
        :param previous_message_df: The message dataframe to use to compute the distances
            (distance from vacuous will be used if this is None)
        :return: The ranked message dataframe sorted by the distance_from_previous column (from high to low)
        """
        messages = self._make_all_messages()
        factors_are_vacuous = [message.factor.is_vacuous for message in messages]
        if all(factors_are_vacuous) and self.verbose:
            print('Warning: All messages are vacuous')
        if len(messages) == 0:
            raise ValueError('no messages to send.')
        message_dict = {'message_object': [], 'message_scope': [], 'sender_id': [], 'receiver_id': [],
                        'distance_from_previous': []}
        for msg in messages:
            message_dict['message_object'].append(msg)
            message_dict['message_scope'].append(msg.var_names)
            message_dict['sender_id'].append(msg.sender_id)
            message_dict['receiver_id'].append(msg.receiver_id)
            if previous_message_df is None:
                # We imagine that vacuous messages were sent at step -1.
                distance = msg.distance_from_vacuous()
            else:
                mask = (previous_message_df['sender_id'] == msg.sender_id) & \
                       (previous_message_df['receiver_id'] == msg.receiver_id)
                previous_msg_list = previous_message_df[mask]['message_object'].values
                assert len(previous_msg_list) == 1, 'Something strange'
                previous_msg = previous_msg_list[0]
                distance = msg.distance_from_other(previous_msg)

            message_dict['distance_from_previous'].append(distance)
        ranked_message_df = pd.DataFrame(message_dict)
        ranked_message_df.sort_values(by='distance_from_previous', ascending=False, inplace=True)

        # debug only
        sorted_index_list = ranked_message_df.sort_values(by='distance_from_previous').index.tolist()

        message_passing_index_string = ''.join([str(i) for i in sorted_index_list])
        self.message_passing_index_strings.append(message_passing_index_string)

        return ranked_message_df

    def _process_graph_async(self, tol, max_iter):
        """
        Perform message passing until convergence (or maximum iterations).
        """
        if len(self._clusters) == 1:
            # The Cluster Graph contains only single cluster. Message passing not possible or necessary.
            self._clusters[0]._factor = self._clusters[0]._factor.reduce(vrs=self.special_evidence.keys(),
                                                                         values=self.special_evidence.values())
            return

        max_message_distance = float('inf')
        previous_message_df = None

        print('Info: Starting iterative message passing.*')
        for iterations in tqdm(range(max_iter), disable=self.disable_tqdm):
            self._conditional_print(f'iteration: {iterations}/{max_iter}')
            if max_message_distance < tol:
                self._conditional_print(f'Info: max_message_distance={max_message_distance} < tol={tol}. Stopping.')
                break

            ranked_message_df = self._get_ranked_message_df(previous_message_df)
            for indx, message_row in ranked_message_df.iterrows():
                message = message_row['message_object']
                distance_from_previous = message_row.iloc[-1]
                if distance_from_previous < 0.0:
                    raise ValueError('Distance from previous message is negative.')
                # TODO: correct this: if the message is not passed, it must not be included in new_message_distances_df
                if distance_from_previous >= tol:
                    if self.debug:
                        print(f'\n {message.sender_id} -> {message.receiver_id}\t\t\t(dfp: {distance_from_previous})')
                        if isinstance(message.factor, Gaussian):
                            message.factor.show(update_covform=False, show_canform=True)
                        else:
                            message.factor.show()
                        self.debug_passed_message_factors.append(message._factor.copy())
                    self._pass_message(message)
                    self.num_messages_passed += 1

            if self.num_messages_passed == 0:
                self._conditional_print('Warning: no messages passed.')
            max_message_distance = ranked_message_df['distance_from_previous'].max()
            previous_message_df = ranked_message_df

            new_message_distances_df = ranked_message_df.loc[:, ranked_message_df.columns != 'message_object']
            if iterations == 0:
                self.message_passing_log_df = new_message_distances_df
            else:
                self.message_passing_log_df = pd.merge(left=self.message_passing_log_df,
                                                       right=new_message_distances_df,
                                                       on=['sender_id', 'receiver_id'],
                                                       suffixes=('', f'_iter_{iterations}'))
        self._conditional_print(f'Info: num_messages_passed = {self.num_messages_passed}')
        if self.make_animation_gif:
            self._make_message_passing_animation_gif()

    def _get_most_informative_message(self, messages):
        """
        Find and remove the most informative message from a list of messages.
        :return: The message with the highest KL between its current and previous iteration (if more than one, only the
         last calculated is returned) and the distance from the previous iteration.
        :rtype: Message, float
        """
        factors_are_vacuous = [message.factor.is_vacuous for message in messages]
        if all(factors_are_vacuous) and self.verbose:
            print('Warning: All messages are vacuous')
        if len(messages) == 0:
            raise ValueError('no messages to send.')

        max_distance = -np.inf
        max_distance_message = None
        distances = []
        for i, msg in enumerate(messages):
            if (msg.sender_id, msg.receiver_id) not in self.last_passed_message_factors_dict:
                # We imagine that vacuous messages were sent at step -1.
                distance = msg.factor.distance_from_vacuous()
            else:
                previous_msg_factor = self.last_passed_message_factors_dict[(msg.sender_id, msg.receiver_id)]
                distance = previous_msg_factor.kl_divergence(msg.factor)
            distances.append(distance)
            if distance > max_distance:
                max_distance_message = msg
                max_distance = distance

        return max_distance_message, max_distance

    def process_graph(self, tol=1e-3, max_iter=50, sync=True):
        """
        Process the graph by performing message passing until convergece.
        :param bool sync: Whether or not to use a synchronous message passing (asynchronous message passing used if False)
        """
        if sync:
            self._process_graph_sync(tol=tol, max_iter=max_iter)
        else:
            self._process_graph_async(tol=tol, max_iter=max_iter)

    # New Synchronous version
    #  TODO: make this more efficient, if no messages have been received by a cluster in the previous round, the next
    #        message iterations from that cluster will be the same.
    def _process_graph_sync(self, tol, max_iter):
        """
        Perform synchronous message passing until convergence (or maximum iterations).
        """
        self.sync_message_passing_max_distances = []
        if len(self._clusters) == 1:
            # The Cluster Graph contains only single cluster. Message passing not possible or necessary.
            self._clusters[0]._factor = self._clusters[0]._factor.reduce(vrs=self.special_evidence.keys(),
                                                                         values=self.special_evidence.values())
            return

        print('Info: Starting iterative message passing.*')
        cluster_messages_dict = self._make_clusters_messages_dict()

        for iterations in tqdm(range(max_iter), disable=self.disable_tqdm):

            self._conditional_print(f'iteration: {iterations}/{max_iter}')
            messages = sum(cluster_messages_dict.values(), [])
            message, distance_from_previous = self._get_most_informative_message(messages)
            self.sync_message_passing_max_distances.append(distance_from_previous)
            if distance_from_previous < tol:
                self._conditional_print(f'Info: distance_from_previous={distance_from_previous} < tol={tol}. Stopping.')
                break
            self._pass_message(message)
            # recompute the messages for the cluster that received the message
            cluster_that_received_msg = self._cluster_id_cluster_dict[message.receiver_id]
            cluster_messages_dict[message.receiver_id] = self._make_cluster_messages(cluster_that_received_msg)
            self.num_messages_passed += 1
            if self.num_messages_passed == 0:
                self._conditional_print('Warning: no messages passed.')

        self._conditional_print(f'Info: num_messages_passed = {self.num_messages_passed}')
        if self.make_animation_gif:
            self._make_message_passing_animation_gif()

    def _pass_message(self, message):
        """
        Pass message to the relevant receiver.
        :param message: The message to pass
        """
        receiver_cluster_id = message.receiver_id
        for cluster in self._clusters:
            if receiver_cluster_id == cluster.cluster_id:
                error_msg = f'Error: {message.var_names} is not a subset of {cluster.var_names}'
                assert set(message.var_names) <= set(cluster.var_names), error_msg
                cluster.receive_message(message)
                self.last_passed_message_factors_dict[(message.sender_id, message.receiver_id)] = message.factor.copy()
                if self.make_animation_gif:
                    cg_animation.add_message_pass_animation_frames(graph=self._graph,
                                                                   frames=self.message_passing_animation_frames,
                                                                   node_a_name=message.sender_id,
                                                                   node_b_name=receiver_cluster_id)

    def _make_message_passing_animation_gif(self):
        print('Making message passing animation.')
        self.message_passing_animation_frames[0].save(fp='./graph_animation.gif',
                                                      format='GIF',
                                                      append_images=self.message_passing_animation_frames[1:],
                                                      save_all=True, duration=400, loop=0)

    @property
    def graph(self):
        return self._graph



"""
A module for factorised factor functionality
"""

from veroku.factors._factor import Factor
from veroku.factors._factor_utils import get_subset_evidence

# TODO: This class in very much a work in progress. This needs to be completed. We must think especially carefully about
#  how we multiply factors when we have special factors such as nonlinear Gaussian factors in the factorised factor.


class FactorisedFactor(Factor):
    """
    A factorised factor class that allows factors to be used in factorised form - a product of other factors.
    This is useful for improving efficiency between factor multiplication and marginalisation. It also allows the
    for a efficient and intuitive representation of independent factors (especially factors with disjoint scopes).
    """

    # TODO: see how this is used in practice and investigate what the best way is to treat the variable name parameters.
    def __init__(self, factors):
        """
        The initialiser.
        :param factors: The list of factors that will be implicitly multiplied.
        """
        assert all([not isinstance(factor, FactorisedFactor) for factor in factors])
        factors_var_name_lists = ([factor.var_names for factor in factors])
        factors_var_name = list(set([var_name for sublist in factors_var_name_lists for var_name in sublist]))
        super().__init__(var_names=factors_var_name)
        self.factors = []
        for factor in factors:
            if isinstance(factor, FactorisedFactor):
                self.factors += [f.copy() for f in factor.factors]
            else:
                self.factors.append(factor.copy())

    @property
    def joint_distribution(self):
        """
        compute the joint distribution.
        :return:
        """
        if len(self.factors) == 0:
            raise ValueError('FactorisedFactor has no factors.')
        joint_distribution = self.factors[0].copy()
        for factor in self.factors[1:]:
            joint_distribution = joint_distribution.multiply(factor)
        return joint_distribution

    def copy(self):
        """
        Copy this factorised factor
        :return: the copied factor
        """
        factor_copies = []
        for factor in self.factors:
            factor_copies.append(factor.copy())
        return FactorisedFactor(factor_copies)

    def multiply(self, factor):
        """
        Multiply in a general factor
        :param factor: any type of factor.
        :return: the resulting Factorised Factor
        """
        # TODO: add preference for factors with conditioning scope? Not sure if practically necessary, but would be best.
        factorised_factor_copy = self.copy()
        index = factorised_factor_copy.first_factor_with_vars_index(factor.var_names)
        if index is not None:
            factorised_factor_copy.factors[index] = factorised_factor_copy.factors[index].multiply(factor)
        else:
            factorised_factor_copy.factors.append(factor)
        return factorised_factor_copy

    def divide(self, factor):
        """
        Divide out a general factor
        :param factor: any type of factor.
        :return: the resulting Factorised Factor
        """
        factorised_factor_copy = self.copy()
        index = factorised_factor_copy.first_factor_with_vars_index(factor.var_names)
        # TODO: fix or remove this. This is breaks when the factors do not have equals methods
        # for i, factorised_comp in enumerate(self.factors):
        #     if factorised_comp.equals(factor):
        #         if len(self.factors) > 1:
        #             del factorised_factor_copy.factors[i]
        #             return factorised_factor_copy
        #         # TODO: add else to make general vacuous factor (wont necessarily be Gaussian)
        if index is not None:
            factorised_factor_copy.factors[index] = factorised_factor_copy.factors[index].divide(factor)
        else:
            raise Exception('Error: Cannot divide factor with disjoint scope.')
        return factorised_factor_copy

    @property
    def is_vacuous(self):
        all_vacuous = all([factor._is_vacuous for factor in self.factors])
        if all_vacuous:
            return True
        none_vacuous = all([not factor._is_vacuous for factor in self.factors])
        if none_vacuous:
            return False
        # TODO: implement this
        raise NotImplementedError()

    def distance_from_vacuous(self):
        # TODO: improve this
        return self.joint_distribution.distance_from_vacuous()

    def kl_divergence(self, factor):
        # TODO: improve this
        return self.joint_distribution.kl_divergence(factor)


    @property
    def num_factors(self):
        """
        Get the number of factors in the factorised factor.
        :return: (int) The number of factors.
        """
        return len(self.factors)

    #def marginalize(self, vrs, keep=False):
    #    """
    #    Marginalise out a subset of the variables in this factor's scope.
    #    :param vrs: (list) the variable names
    #    :param keep: (bool) whether to keep or sum (or integrate) out these variables.
    #    :return: (FactorisedFactor) the resulting marginal
    #    """
    #    cg = ClusterGraph(self.factors, make_animation_gif=False, debug=False, disable_tqdm=True)
    #    cg.process_graph()
    #    #vars_to_keep = super().get_marginal_vars(vrs, keep)
    #    #marginal = cg.get_marginal(vars_to_keep)
    #    posterior_joint = cg.get_posterior_joint()
    #    marginal = posterior_joint.marginalize(vrs, keep=keep)
    #    return marginal

    def marginalize(self, vrs, keep=False):
        """
        Marginalise out a subset of the variables in this factor's scope.
        :param vrs: (list) the variable names
        :param keep: (bool) whether to keep or sum (or integrate) out these variables.
        :return: (FactorisedFactor) the resulting marginal
        """
        # TODO: resolve this: Non-linear Gaussian factors can cause a problem here when there is enough information
        # that they all should be well defined, but the information resides in only one of them for instance. This
        # information takes the form of the observed evidence and the conditioning factors.

        vars_to_keep = super().get_marginal_vars(vrs, keep)
        vars_to_integrate_out_set = set(self.var_names) - set(vars_to_keep)
        factors_in_keep_scope_indices = self.all_intersecting_factors_indices(vars_to_keep)
        if not factors_in_keep_scope_indices:
            raise Exception('cannot marginalize out all variables from FactorisedFactor.')

        factor_marginals = []
        self_copy = self.copy()
        additional_log_weight = 0.0
        self_copy.merge_factors_with_shared_scope(vars_to_integrate_out_set)
        for factor in self_copy.factors:
            factor_vars_to_integrate_out_set = set(factor.var_names).intersection(vars_to_integrate_out_set)
            if factor_vars_to_integrate_out_set == set(factor.var_names):
                additional_log_weight += factor.get_log_weight()
            else:
                factor_vars_to_integrate_out = list(factor_vars_to_integrate_out_set)
                factor_marginal = factor.marginalize(factor_vars_to_integrate_out, keep=False)
                factor_marginals.append(factor_marginal)
        factor_marginals[0]._add_log_weight(additional_log_weight)
        if len(factor_marginals) == 1:
            return factor_marginals[0]
        marginal_factor = FactorisedFactor(factor_marginals)
        return marginal_factor

    # TODO: remove this
    def marginalise_OLD(self, vrs, keep=False):
        """
        Marginalise out a subset of the variables in this factor's scope.
        :param vrs: (list) the variable names
        :param keep: (bool) whether to keep or sum (or integrate) out these variables.
        :return: (FactorisedFactor) the resulting marginal
        """
        vars_to_keep = super().get_marginal_vars(vrs, keep)
        factors_in_keep_scope_indices = self.all_intersecting_factors_indices(vars_to_keep)
        factors_in_nokeep_scope_indices = list(set(range(self.num_factors)) - set(factors_in_keep_scope_indices))
        if not factors_in_keep_scope_indices:
            raise Exception('cannot marginalize out all variables from FactorisedFactor.')

        keep_scope_factor_product = self.factors[factors_in_keep_scope_indices[0]].copy()
        for index in factors_in_keep_scope_indices[1:]:
            factor_component = self.factors[index]
            keep_scope_factor_product = keep_scope_factor_product.multiply(factor_component)

        # TODO: this can potentially be improved by checking if all the keep scope factors maybe have
        #  set(vars)==vars_to_keep -> then we do not even need to multiply them together.
        if isinstance(keep_scope_factor_product, FactorisedFactor):
            raise TypeError
        keep_scope_factor_product_marginal = keep_scope_factor_product.marginalize(vars_to_keep, keep=True)

        joint_for_full_integration = self.factors[factors_in_nokeep_scope_indices[0]].copy()
        #TODO: here we can also optimise be integrating factors with disjoint variable scopes separately.
        for index in factors_in_nokeep_scope_indices[1:]:
            joint_for_full_integration = joint_for_full_integration.multiply(self.factors[index])
        additional_log_weight = joint_for_full_integration.get_log_weight()
        keep_scope_factor_product_marginal.add_log_weight(additional_log_weight)

        return keep_scope_factor_product_marginal

    #TODO: make these more efficient by storing joint_distribution
    def get_cov(self):
        return self.joint_distribution.get_cov()

    def get_mean(self):
        return self.joint_distribution.get_mean()

    # TODO: uncomment and test
    #def merge_dependent_factors(self):
    #    """
    #    Merge all factors (by multiplying) that have overlapping scopes.
    #    This is useful for?
    #    :return:
    #    """
    #    merged_factors = []
    #    already_merged_factor_indices = []
    #    for i, factor_i in enumerate(self.factors):
    #        factor_i_merged = factor_i.copy()
    #        for j in range(i+1,len(self.factors)):
    #            factor_j = self.factors[j]
    #            if j not in already_merged_factor_indices:
    #                if set(factor_i.var_names).intersection(set(factor_j.varnames)) > 0:
    #                    factor_i_merged = factor_i_merged.multiply(factor_j)
    #                    already_merged_factor_indices.append(j)
    #        merged_factors.append(factor_i_merged)
    #    return FactorisedFactor(merged_factors)

    def merge_factors_with_shared_scope(self, vrs_set):
        """
        Merge all factors in groups that all contain variables in vrs and have overlap within these subsets.
        This function will typically be used to ensure that all factors containing variables that need to be marginalised
        out are merged together into disjoint (in terms of the integrand variables) factors, so that these larger factors
        can be marginalised independently.
        :param vrs_set: (string list) The common scope variables to be considered to determine 'dependent' factors.
        """
        observed_factor_vars_list = []

        for f in self.factors:
            observed_factor_vars = set(f.var_names).intersection(vrs_set)
            observed_factor_vars_list.append(observed_factor_vars)

        merged_factors = []
        merged_factors_indices = []
        for i, observed_factor_i_vars in enumerate(observed_factor_vars_list):
            if i not in merged_factors_indices:
                merged_factor = self.factors[i]
                for j in range(i + 1, len(observed_factor_vars_list)):
                    observed_factor_j_vars = observed_factor_vars_list[j]
                    if j not in merged_factors_indices:
                        if len(observed_factor_i_vars.intersection(observed_factor_j_vars)) > 0:
                            merged_factor = merged_factor.multiply(self.factors[j])
                            merged_factors_indices.append(j)
                merged_factors.append(merged_factor)
        self.factors = merged_factors

    def reduce(self, vrs, values):
        """
        Observe certain values of the variables in vrs.
        :param vrs: (list) The observed variables.
        :param values: (list) Their values.
        :return: (FactorisedFactor) the reduced factor
        """
        all_evidence_dict = dict(zip(vrs, values))
        reduced_factors = []
        additional_log_weight = 0.0
        for factor in self.factors:
            factor_var_names = factor.var_names
            if set(vrs) == set(factor_var_names):
                additional_log_weight += factor.log_potential(x_val=values, vrs=vrs)
            elif len(set(vrs).intersection(set(factor_var_names))) > 0:
                subset_vrs, subset_values = get_subset_evidence(all_evidence_dict=all_evidence_dict,
                                                                subset_vars=factor.var_names)
                reduced_factor = factor.reduce(subset_vrs, subset_values)
                reduced_factors.append(reduced_factor)
            else:
                reduced_factors.append(factor.copy())
        adjusted_weight_factor_0 = reduced_factors[0]
        adjusted_weight_factor_0._add_log_weight(additional_log_weight)
        reduced_factors[0] = adjusted_weight_factor_0
        return FactorisedFactor(reduced_factors)

    def first_factor_with_vars_index(self, var_names):
        """
        returns index of the first factor in self_factors with var_names in its scope.
        """
        # TODO: add check for equals variable names - will be faster for multiplication
        for i, factor in enumerate(self.factors):
            if set(var_names).issubset(set(factor.var_names)):
                return i
        return None

    def all_intersecting_factors_indices(self, var_names):
        """
        return indices of factors in self_factors which has overlap with var_names.
        """
        indices = []
        # TODO: add check for equals variable names - will be faster for multiplication
        for i, factor in enumerate(self.factors):
            if len(set(var_names).intersection(set(factor.var_names))) > 0:
                indices.append(i)
        return indices

    def equals(self, factor):
        """
        Check if this factor is the same as another factor.

        :param factor: The other factor to compare to.
        :type factor: FactorisedFactor
        :return: The result of the comparison.
        :rtype: bool
        """
        if not isinstance(factor, FactorisedFactor):
            raise NotImplemented()
        if set(factor.var_names) != set(self.var_names):
            return False
        if len(self.factors) != len(factor.factors):
            return self.joint_distribution.equals(factor.joint_distribution)

        self_comps_with_cpoies_in_factor = []
        for i, self_factor_i in enumerate(self.factors):
            if i not in self_comps_with_cpoies_in_factor:
                for j, other_factor_j in enumerate(factor.factors):
                    if self_factor_i.equals(other_factor_j):
                        self_comps_with_cpoies_in_factor.append(i)
        if len(self_comps_with_cpoies_in_factor) == len(self.factors):
            return True
        return False

    def show(self):
        """
        Display this factorised factor
        """
        for i, factor in self.factors:
            print(f'\n factor {i}/{len(self.factors)}:')
            factor.show()
